package abstractPracticeTheory;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import repast.simphony.context.Context;
import repast.simphony.context.DefaultContext;
import repast.simphony.context.space.grid.GridFactory;
import repast.simphony.context.space.grid.GridFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.query.space.grid.MooreQuery;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridBuilderParameters;
import repast.simphony.space.grid.GridDimensions;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.space.grid.MooreContains;
import repast.simphony.space.grid.SimpleGridAdder;
import repast.simphony.space.grid.WrapAroundBorders;
import repast.simphony.util.SimUtilities;
import repast.simphony.util.collections.IndexedIterable;

public class GridBuilder extends DefaultContext<Object> implements ContextBuilder<Object> {
	
	public static final int NR_BITS = 50;
	public static final int gridXSize = 30;
	public static final int gridYSize = 30;
	
	//max 5 practices allowed
	public static Color[] materialColors = {Color.blue, Color.green, Color.yellow, Color.PINK, Color.red};
	
	private Grid <Object > grid;
	
	//For data collection
	//Counts total number of changes in material for all individuals over all steps
	protected int totalChangesMaterial = 0;
	//store routine behaviour of last 50 steps to calculate moving average
	protected int [] routineLastNSteps;
	//store MoransI of last N steps to calculate moving average
	protected double [] moransIGreen;
	protected double [] moransIBlue;
	protected double [] moransIPink;
	protected double [] moransIRed;
	protected double [] moransIYellow;
	
	public GridBuilder(){
		super();
	}
	
	
	@Override
	public Context build(Context<Object> context) {
		
		context.setId("abstractPracticeTheory");
		
		GridFactory gridFactory = GridFactoryFinder.createGridFactory (null);
		grid = gridFactory.createGrid ("grid", context ,
		new GridBuilderParameters <Object >( new WrapAroundBorders (),
		new SimpleGridAdder <Object >() ,
		true , gridXSize, gridYSize));
	
		//get number of GridCells
		GridDimensions gd = grid.getDimensions();
		int cellCount = gd.getHeight()*gd.getWidth();
		ArrayList<Individual> inds = new java.util.ArrayList<Individual>(cellCount);
		ArrayList<ColoredMaterial> materials = new java.util.ArrayList<ColoredMaterial>(cellCount);
		Individual ind;
		
		//check GUI for number of practices and types of individuals
		Parameters params = RunEnvironment.getInstance().getParameters();
		int nrIndTypes = (Integer)params.getValue("nrIndTypes");
		int nrMaterials = (Integer) params.getValue("nrMaterial");
		
		//create different practices
		ColoredMaterial[] matA = new ColoredMaterial[nrMaterials];
		for(int i=0; i<nrMaterials;i++){
			matA[i] = new ColoredMaterial(NR_BITS,materialColors[i]);
		}
		//create meaning and skills for types of individuals
		Meaning[] mA = new Meaning[nrIndTypes];
		Competence[] sA = new Competence[nrIndTypes];
		for(int i=0; i<nrIndTypes;i++){
			mA[i] = new Meaning(NR_BITS);
			mA[i].setUpRandom();
			sA[i] = new Competence(NR_BITS);
			sA[i].setUpRandom();
		}
		//create (resp. clone) and connect practice and individual
		ColoredMaterial mat;
		int indType;
		int x=0, y=0;
		for(int i=0;i< cellCount; i++){ 
			if(RandomHelper.nextDouble()< ((Integer)params.getValue("percentOccupiedCells")/100.0)){		//share of cells is occupied
				mat = matA[RandomHelper.nextIntFromTo(0, nrMaterials-1)].cloneThis();
				if(nrIndTypes>0){			//randomly choose one type of individual 
					indType = RandomHelper.nextIntFromTo(0, nrIndTypes-1);
					ind = new Individual(mA[indType].cloneThis(),sA[indType].cloneThis(),mat); //clone meaning and skills so they can be changed for each individual separately
				}
				else{//create random Individual
					Meaning m = new Meaning(NR_BITS);
					m.setUpRandom();
					Competence s = new Competence(NR_BITS);
					s.setUpRandom();
					ind = new Individual(m,s,mat);
				}
				inds.add(ind);
				materials.add(mat);
				//move to Grid
				context.add(ind);
				context.add(mat);
				ind.setGrid(grid);
				mat.setGrid(grid);
				grid.moveTo(ind, x, y);
				grid.moveTo(mat, x, y);
			}
			x=(x+1)%gd.getWidth();
			if(x==0)y++;
		}
		
		//set fixed simulation length if respective parameter is set in the GUI
		
		Object o; 
		try{
			o = params.getValue("runLength");
		}catch(Exception e){o = null;};
		
		if((o !=null)&&((Integer)o > 0)){
			RunEnvironment.getInstance().endAt((Integer) o); //set length of simulation runs to value defined in GUI (for usage in Parameter Sweep)
		}
		
		return context ;
	}
	
	///////////////////////
	///Data collection (for visualization of some results/////////
	/////////////////////
	//low priority: do this after all individuals have executed all methods
	
	@ScheduledMethod(start=1, interval=1, shuffle=true, priority=0.5)
	public void collectData(){
		totalChangesMaterial+=countChangesMaterial();
		int tick = (int) RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		
		if(routineLastNSteps == null){
			//create array to store steps for later calculation of moving avg
			routineLastNSteps = new int[50];
		}
		routineLastNSteps[tick % routineLastNSteps.length] = this.getNrIndividualsRoutineBehaviour();
		
		if(moransIGreen == null){
			//create array to store steps for later calculation of moving avg
			moransIGreen = new double [50];
		}
		moransIGreen[tick % moransIGreen.length] = getMoransIGreen();

		if(moransIBlue == null){
			//create array to store steps for later calculation of moving avg
			moransIBlue = new double [50];
		}
		moransIBlue[tick % moransIBlue.length] = getMoransIBlue();
		
		if(moransIPink == null){
			//create array to store steps for later calculation of moving avg
			moransIPink = new double [50];
		}
		moransIPink[tick % moransIPink.length] = getMoransIPink();
		
		if(moransIRed == null){
			//create array to store steps for later calculation of moving avg
			moransIRed = new double [50];
		}
		moransIRed[tick % moransIRed.length] = getMoransIRed();
		
		
		if(moransIYellow == null){
			//create array to store steps for later calculation of moving avg
			moransIYellow = new double [50];
		}
		moransIYellow[tick % moransIYellow.length] = getMoransIYellow();
		
	}
	
	protected int countChangesMaterial(){
		
		int count =0;
		IndexedIterable<Object> inds = getObjects(Individual.class);
		for(int i=0; i<inds.size(); i++){
			if( ((Individual)inds.get(i)).bestMaterial != null){	//this indicates that this individual has chosen to change material in this step; works if this method is called after Individual.postStep()
				count ++;
			}
		}
		return count;
	}
	
	
	
	
	////////////////////////////////
	//// VISUALIZATION ////////////
	///////////////////////////////
	
	
	//////////////////////////////////////
	///Changes Material /////////////////
	public int getTotalChangesMaterial(){
		return totalChangesMaterial;
	}
	public String getNameDataSetTotalChangesMaterial(){
		return "total changes of Material";
	}
	
	public double getShareIndividualsWhichChangedMaterial(){
		
		int count =0;
		Iterator<Object> indIt = this.getObjects(Individual.class).iterator();
		while(indIt.hasNext()){
			if(((Individual)indIt.next()).bestMaterial != null){	//bestPractice is used by Individual to store a Practice that will be adopted in the same step. If not null this means there has been a change to Material.
				count++;
			}
		}
		return count;
	}
	public String getNameDataSetShareIndividualsWhichChangedMaterial(){
		return "Share of Individuals who changed Material in this step.";
	}
	
	
	
	//////////////////////////////////////
	///Spread////////////////////////////////
	
	/*Returns the share of Individuals which have adopted the most dominant practice*/
	public double getMaxSpread(){
		
		double maxSpread=0.0, s;
		for(int c=0;c<materialColors.length;c++){
			s= getSpread(materialColors[c]);
			if(s>maxSpread) maxSpread =s;
		}
		return maxSpread;
	}
	
	public String getNameDataSetMaxSpread(){
		return "spread of dominant practice";
	}
	
	protected double getSpread(Color c){
		
		int count =0;
		IndexedIterable<Object> materials = getObjects(ColoredMaterial.class);
		for(int i=0; i<materials.size(); i++){
			if(((ColoredMaterial)materials.get(i)).color.equals(c)){
				count++;
			}
		}
		return ((double) count)/this.getObjects(Individual.class).size();
	}
	
	public double getSpreadGreen(){
		return getSpread(Color.green);	
	}
	public double getSpreadBlue(){
		return getSpread(Color.blue);	
	}
	public double getSpreadYellow(){
		return getSpread(Color.yellow);	
	}
	public double getSpreadPink(){
		return getSpread(Color.pink);	
	}
	public double getSpreadRed(){
		return getSpread(Color.red);	
	}
	
	//////////////////////////////////////
	////Clustering///////////////////////////
	
	
	//Clustering of Material Bits
	//Summiere fr alle Material Xi und ihre Nachbar Xj: 
	//(Xi*Xj/Xi.length-0.5) mit Xi*Xj ist Art Vektorprodukt (summiere ber Bits: bereinstimmende Bits=1, sonst 0), normiere mit length 
	//Normiere Gesamtclustering mit N*8*(1.0-0.5): alle Moore-Nachbarn htten perfekte bereinstimmung
	//--> Wert in [-1,1]: -1 stark anti-korrelliert, 0~Zufall, 1~starke Korrelation
	//Zufallsbesetzung mit best Anzahl Material n -> 1/n Nachbarn haben identisches Material -> clustering >0
	public double clusteringMaterialBits(){
		
		//Assume fully filled Grid
		
		int N=this.getObjects(Individual.class).size();
		Grid grid = (Grid) this.getProjection("grid");
		
		double sum=0.0; 
		
		//fr alle Zellen 
		for(int x=0; x<gridXSize; x++){
			for(int y=0; y<gridYSize; y++){
				ColoredMaterial cm = null;			//get Material at the location
				Iterator<Object> itCell = grid.getObjectsAt(x,y).iterator();
				while(itCell.hasNext()){
					Object o = itCell.next();
					if(o instanceof ColoredMaterial){
						cm = (ColoredMaterial) o;
					}
				}
				//go over Moore neighbours and add to sum 
				ColoredMaterial neighbour = null;
				for(int dx=-1; dx<=1; dx++){
					for(int dy=-1; dy<=1; dy++){
						if(! (dx==0 && dy==0)){		//exclude center cell
							//get Material at neighbour cell
							Iterator<Object> itNCell = grid.getObjectsAt(x+dx,y+dy).iterator(); //assume that Grid handles Torus behaviour 
							while(itNCell.hasNext()){
								Object o = itNCell.next();
								if(o instanceof ColoredMaterial){
									neighbour = (ColoredMaterial) o;
								}
							}
							//count similar bits
							double similarBits =0.0;
							for(int b=0;b<cm.length(); b++){ //go over all bits
								if(cm.get(b)==neighbour.get(b)){
									similarBits+=1.0;
								}
							}
							//normiere und ziehe erwartete zufllige bereinstimmung ab
							sum += (similarBits/cm.length()-0.5);
						}
					}
				}
			}
		}
		
		double clustering = sum/(8*N*(1.0-0.5)); 
		return clustering;
	}
	
	//Calculation of Moran's I
	public double getMoransI(Color c){
		
		//Assume fully filled Grid
		//calculation according to http://en.wikipedia.org/wiki/Moran's_I
		
		int N=this.getObjects(Individual.class).size();
		double spread = getSpread(c);
		
		if(spread > 0){
		
			double xbar = ((double)spread)/N; //1=Material hat color c, 0=ein anderes --> avg = spread/N
			double denom = spread*N*(1-xbar)*(1-xbar) + (N - spread*N)*(0-xbar)*(0-xbar);//SUMi (Xi-xbar)^2
			double sumWeights = N*8; //SUMij Wij (each cell has 8 Moore-Neighbours)
			double sum = 0.0; //to calculate SUMij Wij (Xi-xbar)(Xj-xbar) in the following
			
			Grid grid = (Grid) this.getProjection("grid");
			
			for(int x=0; x<gridXSize; x++){
				for(int y=0; y<gridYSize; y++){
					ColoredMaterial cm = null;			//get Material at the location
					Iterator<Object> itCell = grid.getObjectsAt(x,y).iterator();
					while(itCell.hasNext()){
						Object o = itCell.next();
						if(o instanceof ColoredMaterial){
							cm = (ColoredMaterial) o;
						}
					}
					//go over Moore neighbours and add to sum 
					ColoredMaterial neighbour = null;
					for(int dx=-1; dx<=1; dx++){
						for(int dy=-1; dy<=1; dy++){
							if(! (dx==0 && dy==0)){		//exclude center cell
								//get Material at neighbour cell
								Iterator<Object> itNCell = grid.getObjectsAt(x+dx,y+dy).iterator(); //assume that Grid handles Torus behaviour 
								while(itNCell.hasNext()){
									Object o = itNCell.next();
									if(o instanceof ColoredMaterial){
										neighbour = (ColoredMaterial) o;
									}
								}
								//wij = 1 (because Moore-Neighbours)
								int xi=0, xj=0;
								if(cm.color.equals(c)){
									 xi = 1;
								}
								if(neighbour.color.equals(c)){
									xj = 1;
								}
								sum += (xi-xbar)*(xj-xbar);
							}
						}
					}
				}
			}
			
			double moransI = N/sumWeights*sum/denom; 
			return moransI;
		}
		else{ //spread=0
			return 0.0;
		}
	}
	
	
	public double getMoransIGreen(){
		return getMoransI(Color.green);	
	}
	public double getMoransIBlue(){
		return getMoransI(Color.blue);	
	}
	public double getMoransIYellow(){
		return getMoransI(Color.yellow);	
	}
	public double getMoransIPink(){
		return getMoransI(Color.pink);	
	}
	public double getMoransIRed(){
		return getMoransI(Color.red);	
	}
	
	public double getMoransIGreenMovingAvg(){
		return getAvg(moransIGreen);	
	}
	public double getMoransIBlueMovingAvg(){
		return getAvg(moransIBlue);	
	}
	public double getMoransIYellowMovingAvg(){
		return getAvg(moransIYellow);		
	}
	public double getMoransIPinkMovingAvg(){
		return getAvg(moransIPink);		
	}
	public double getMoransIRedMovingAvg(){
		return getAvg(moransIRed);		
	}
	
	
	//Returns average of array
	public double getAvg(double[] a){
		
		if(a!=null && RunEnvironment.getInstance()!= null && RunEnvironment.getInstance().getCurrentSchedule().getTickCount() > a.length){
		
				double sum = 0.0; 
				for(int i=0;i<a.length;i++){
					sum+=a[i];
				}
				return sum/a.length;
			
		}
		else{
			return Double.NaN;
		}
	}
	
	//////////////////////////////////////
	//////Routine behaviour/////////////////
	
	public int getNrDeliberateChoices(){
		int count =0;
		Iterator<Object> indIt = this.getObjects(Individual.class).iterator();
		while(indIt.hasNext()){
			if(((Individual)indIt.next()).getDeliberateChoice()){
				count++;
			}
		}
		return count;
	}
	public String getNameDataSetNrIndividualsDiscursiveConsciousnessActive(){
		return "Nr Individuals actively looking for practices";
	}
	
	public int getNrIndividualsRoutineBehaviour(){
		//all individual minus those whose discursiveConsciousness is activated
		return this.getObjects(Individual.class).size() - getNrDeliberateChoices();
	}
	public String getNameDataSetNrIndividualsRoutineBehaviour(){
		return "Nr Individuals with routine behaviour";
	}
	
	//Returns share of individuals with routine behaviour, moving average over last 50 steps
	public double getRoutineBehaviourMovingAvg(){
		
		if(routineLastNSteps!= null && RunEnvironment.getInstance()!= null && RunEnvironment.getInstance().getCurrentSchedule().getTickCount() > routineLastNSteps.length){
		
				double sum = 0.0; 
				for(int i=0;i<routineLastNSteps.length;i++){
					sum+=routineLastNSteps[i];
				}
				return sum/routineLastNSteps.length/this.getObjects(Individual.class).size();
			
		}
		else{
			return Double.NaN;
		}
	}
	
	///////////////
	////Moves//////
	
	public int getNrMovesInTimeStep(){
		int count =0;
		Iterator<Object> indIt = this.getObjects(Individual.class).iterator();
		while(indIt.hasNext()){
			if(((Individual)indIt.next()).getMoved()){
				count++;
			}
		}
		return count;
	}
	public String getNameDataSetNrMovesInTimeStep(){
		return "Nr of individuals who moved in this step";
	}
	
	
	//////////////
	//Init state///
	
	public int getInitialNrPractices(){
		return (Integer) RunEnvironment.getInstance().getParameters().getValue("nrMaterial");
	}
	public int getSocialInfluence(){
		return (Integer) RunEnvironment.getInstance().getParameters().getValue("socialInfluence");
	}
	public int getIndividualTypes(){
		return (Integer) RunEnvironment.getInstance().getParameters().getValue("nrIndTypes");
	}
	public double getHabitAcquisitionSpeed(){
		return (Double) RunEnvironment.getInstance().getParameters().getValue("ha");
	}
	
	
	
	public String getNameDataSetClustering(){
		return "Clustering";
	}
	
}
